(* ===================================================================

  This file defines a syntactical embedding of L0 in HOL.

  L0 a minimalistic imperative language. In some places the syntax is
  still very general, which allows some room for specializations.

  author: wishnu prasetya

=================================================================== *)

local

open Parse HolKernel boolLib bossLib 
     stringTheory optionTheory integerTheory ;

in

(*

load "stringTheory" ;
load "optionTheory" ;
load "integerTheory" ;
open stringTheory optionTheory integerTheory ;

*)

val _ = new_theory "L0" ;

(* ===================================================================
  
SYTAX of L0

// Type : (underspecified) any HOL-type

  Type --> HOL-Type

// Program expression : (underspecified) any HOL-term

  Expr --> HOL-term

// Instruction (statemet)
  
  Instr -->
        skip
 
     |  print (Expr)           // printing to screen 

     |  Expr /:= Expr          // represents assignment

     |  /{ i ; ... ; j /}      // sequence of instructions

     |  /@ ProgName (ActualParams)   // call a program

     |  Expr /@= ProgName(ActualParams)
                              //  call a program, then assign the return value to x

     |  if (Expr) then /{ Instr /} else /{ Instr /}
                             // usual meaning

     |  let 
         Var = Expr and 
         ...
         Var = Expr
         in /{ Instr /}    // introducing local variables, initialized.
                           // note: "let" is misleading; but I don't know
                           // how to override the parser so that "var"
                           // can be used instead.

    |  while (Expr) wdo /{ Instr /}

                           // if the first instruction in the body of
                           // while is an assert i, the i will be interpreted
                           // as specifying an invariant for the loop

    |  assert (Assertion)  // Assert specifies a predicate that must hold
                           // at that point. The intended use of "assert"
                           // is for verification (or testing) purpose only.
                           // Its normal executional semantic is just a skip.


NOTE: no exception is supported at the moment.

// Program Definition

  ProgDef -->
         ProgName (FormalParams) = ProgDefRHS

  NOTE
  (1) Program definitions will be introduced via the standard HOL mechanisms 
      for defining constants. This let us reuse HOL's own infrastructure
      to provide basic management of LO programs.

  (2) L0 programs are fully modular. That is, they are not allowed 
      to access a global variable unless it is passed as a parameter.

  ProgDefRHS --> pre   (Assertion)
                 post  (AssertionRet)   // Same as assertion, but here you can use
                                           the "ret" keyword 
                                           Also note that reference to pass-by-val
                                           parameters in "post" refer to their
                                           initial values (since the program is
                                           supposed to have no effect on them)

                 do   ProgBody


   ProgBody --> /{ Instr /} return (Expr)

             |  let Var = Expr and 
                    ...
                    Var = Expr
                in
                ProgBody

// Parameters

  ActualParams --> ActualPar, ... , ActualPar
  ActualPar    --> REF var    // var to be passed by reference
                 | Expr       // to be passed by value


  FormalParams --> FormalPar, ... , FormalPar
  FormalPar    --> REF var    // pass-by-ref param
                   var        // pass-by-value param
  

// Assertion: (underspecified)

   Assertion  -->  old var    // refers to the "initial" value of "var";
                                 if "var" is the program parameter then
                                 old var is the value of var when it is
                                 passed to the program;
                                 if it is a local variable then old var
                                 is the value of var when it is initialized
                                 in its declaration.
                                 (!) Notice that the argument of old
                                     can only be a var!
                | HOL-term

  AssertionRet --> same as Assetion +  "ret"        

                   // "ret' refers to return value; can only
                      occur in the post-cond of a program.
             

=================================================================== *)


(* ===================================================================

   Introducing new types to represent various syntactic categories
   of L0.

=================================================================== *)

(* --- Instruction *)

val _ = new_type ("INSTR",0) ;

(* --- Program  
  
   An L0 program that returns a value of type "r" will be represented
   by a value of type "r PROG" *)

val _ = new_type ("PROG",1)  ;

(* --- ProgBody  
  
   A ProgBody is the top-level instruction in the body of a program.
   It is represented by the type "r PBODY" where r is the type of
   the value returned within the ProgBody. *)

val _ = new_type ("PBODY",1)  ;

(*--- use "void" below to represent empty tuple of params *)

val _ = Hol_datatype `void = void | none` ;

(* --- pass-by-ref parameter *)

val _ = Hol_datatype `REF = REF of 'a` ;


(* ===================================================================

   Introducing various constants and the corresponding syntax 
   embedding of L0.

=================================================================== *)


(* --- print *)

val _ = new_constant ("printInt", Type `:int->INSTR`) ;
val _ = new_constant ("printStr", Type `:string->INSTR`) ;
val _ = overload_on ("print", Term `printInt`) ;
val _ = overload_on ("print", Term `printStr`) ;

(* --- assignment *)

val _ = add_infix    ("/:=", 31, HOLgrammars.NONASSOC) ;
val _ = new_constant ("/:=", Type `:'a->'a->INSTR`) ;

(* --- sequence of instructions 

       Allowing them to take a more general type so that the syntax
       can be reused else where. *)

val _ = new_constant ("skip", Type `:'a`) ;

val _ = new_constant ("seq", Type `:'a->'a->'a`) ;

val _ = add_listform {separator  = [TOK ";", BreakSpace(1,0)], 
                      leftdelim  = [TOK "/{"], 
                      rightdelim = [TOK "/}"],
                      cons = "seq", nilstr = "skip"};

(* --- program call *)

val _ = new_constant ("/@", Type `:('params->'retval PROG)->'params->INSTR`) ;

val _ = add_infix    ("/@=", 34, HOLgrammars.NONASSOC) ;
val _ = new_constant ("/@=", Type `:'retval->'retval PROG->INSTR`) ;

(* --- if-then_else: using HOL's built-in *)

(* --- let: using HOL's built-in *)

(* --- while *)

val _ = new_constant ("wloop", Type `:bool -> INSTR -> INSTR`) ;

val _ = add_rule{term_name = "wloop", fixity = TruePrefix 19,
                 pp_elements = [PPBlock([TOK "while", BreakSpace(1,0),
                                         TM, BreakSpace(1,0),
                                         TOK "wdo",
                                         BreakSpace(1,0) ], 
                                        (PP.CONSISTENT,3)),
                                BeginFinalBlock(PP.CONSISTENT,0) ],
                paren_style = OnlyIfNecessary,
                block_style = (AroundEachPhrase,(PP.CONSISTENT,0))};

(* --- assert *)

val _ = new_constant ("assert", Type `:bool -> INSTR`) ;

(* --- return-statement within a Prog-body *)

val _ = add_infix    ("return", 29, HOLgrammars.NONASSOC) ;
val _ = new_constant ("return", Type `:INSTR -> 'retval -> 'retval PBODY`) ;


(*--- RHS of program definition *)

val _ = new_constant ("progdef_rhs", Type `:bool     (* pre *)
                                            ->bool   (* post *)
                                            ->'retval PBODY
                                            ->'retval PROG`) ;

val _ = add_rule{term_name = "progdef_rhs", fixity = TruePrefix 21,
                 pp_elements = [PPBlock([TOK "pre", HardSpace 1,
                                         TM, BreakSpace(1,0),
                                         TOK "post", HardSpace 1,
                                         TM, BreakSpace(1,0),
                                         TOK "do",  HardSpace 1 ], 
                                        (PP.CONSISTENT,0)),
                                BeginFinalBlock(PP.CONSISTENT,3) ],
                paren_style = OnlyIfNecessary, 
                block_style = (AroundEachPhrase,(PP.CONSISTENT,0))};

(*--- assertion *)

(* return value *)
val _ = new_constant ("ret", Type `:'a`) ;

(* pre-condition value of a variable *)
val old_def = new_constant ("old", Type `:'a->'a`) ;

(* ===================================================================

   Some examples of an L0 program we can now write in HOL.

   Note that HOL will parse and type-check the program for 
   you :)

   Define 

      `swap (REF x,REF y) 
          =
          pre  T
          post (x = old y) /\ (y = old x)
          do 
          let tmp = 0
          in /{ tmp /:= x   ;
                x   /:= y   ;
                y   /:= tmp   /}

          return void
      ` ;

 
  Define `sumseq (n) =

              pre  (0<=n)
              post (ret = (n*n - n)/2)

              do
              let i = 0 and s = 0 in
              /{ while (i<n) wdo
                    /{ assert ((s = (i*i - i)/2) /\ 0<=i /\ i<=n )
                     ; s /:= s+i
                     ; i /:= i+1 /}
              /}
              return s
          ` ;



=================================================================== *)

val _ = export_theory();

end ; 
