(* ===================================================================

   VCG, represented as a functor

=================================================================== *)

local 

open Parse HolKernel boolLib 

open Option Err 
open L0Theory L0_emb_mldata L0_emb_basic 
open Logic_Util Logic 

infixr 5 AND OR ;
infixr 4 IMP ;
infixr   //  ;
infix    <-  ;

fun snd (_,y) = y ;
fun fst (x,_) = x ;

fun listUpdate (i,f) s 
    =
    let
    fun update z _ []     = List.rev z
      | update z i (x::s) = if i=0 
                            then List.rev z @ (f x::s)
                            else update (x::z) (i-1) s
    in
    if i<0 then s
    else update [] i s
    end 


fun zipWith f s t = List.map f (ListPair.zip (s,t))

in



functor build_VCG (L : Logic) =
   struct

   local

   fun printlist []     = ()
     | printlist [s]    = print s
     | printlist (s::z) = (print s; print ", "; printlist z)
   fun println s = (print "\n"; print s)

   fun verified (_,r) = isSome r
   val unverified     = not o verified

   val WP  = fold_INSTR L.wpAlgebra

   fun printError e 
         =
         if L.isErrorEchoOn() then printL0Error e else ()


   in


   (* --- 
     a mutable variable to hold the VCs that results from the reduction 
   --- *)

   val (VCs : ( string         (* a name for each group of VCs             *)
              * ( VC           (* individual verification condition        *)
                * thm option   (* NONE if still unverified, else SOME thm  *)
                ) list
              ) 
              list
              ref  )
       =
       ref [] ;

   (* --- a function to obtain the status of verification so far --- *)
   fun status () =
         let
         fun getVClabel (vc,marker)  = 
             let
             val lab = #VClabel vc
             in
             if lab = "" then ".-" else "." ^ lab
             end
         fun getVerified (groupName,s)  
             = 
             List.map (fn vc => groupName ^ getVClabel vc) (List.filter verified s)
         fun getUnverified (groupName,s)  
             = 
             List.map (fn vc => groupName ^ getVClabel vc) (List.filter unverified s)
         val verifiedVCs
             =
             List.concat (List.map getVerified (!VCs))
         val unverifiedVCs
             =
             List.concat (List.map getUnverified (!VCs))
         in
         if null unverifiedVCs then (println "STATUS: DONE"; println "" )
         else
         ( println "STATUS:"
         ; println ("VERIFIED (" ^ Int.toString (length verifiedVCs) ^ "):")
         ; println "   " ; printlist verifiedVCs
         ; println ("UNVERIFIED (" ^ Int.toString (length unverifiedVCs) ^"):")
         ; println "   " ; printlist unverifiedVCs
         ; println "" )
         end 
			 

   (* --- a function to collect the resulting theorems --- *)
   fun conclude () =
         let
         fun getThm groupName (vc,marker)  = 
             let
             val lab  = #VClabel vc
             val lab_ = groupName ^ (if lab = "" then  ".-" else "." ^ lab)
             in
             (lab_, valOf marker)
             end
         fun getVerified (groupName,s)  
             = 
             List.map (getThm groupName) (List.filter verified s)
         fun getUnverified (groupName,s)  
             = 
             List.map (fn vc => ()) (List.filter unverified s)
         val verifiedVCs
             =
             List.concat (List.map getVerified (!VCs))
         val unverifiedVCs
             =
             List.concat (List.map getUnverified (!VCs))
         val error = { errorCode   = SOME "conclude"
                     , errorWhere  = NONE
                     , errorInside = NONE
                     , errorWhy    = SOME "There are still some unverified VCs"  } ;
         in
         if null unverifiedVCs
         then verifiedVCs
         else (printError error; raise L0Exn error)
         end


   (* ===================================================================

      The reduce function: to reduce the specification of a program.

      This is a bit complicated. Consider a specification:

         Prog(r,v) = pre  P 
                     post Q 
                     /{ body /} return e

      (!!) note: reference to v in Q refer to the initial value of v

      The rule to such a spec at the top-level is given below:
              
              let w = WP body (Q[old v/v])
              (P ==> w)[r/old r, v/old v]
               generated VCs[r/old r, v/old v]
              ----------------------------------
              {P} body__e  {Q}
       
       where body__e = /{ body ; ret := e /}


   =================================================================== *)


   fun VCG (i:INSTR_) (post:term) (pre:term) finalsubst
                =
                ( L.initializeState() 
                ; let
                  val w = (WP i post
                           handle
                           L0Exn e => (printError e; raise L0Exn e))
                  in 
                  L.add_VCs [ { VClabel = "init", 
                                VC      = pre IMP w
                              } ] 
                  end
                ; List.map (fn v => ( { VClabel = #VClabel v
                                      , VC = finalsubst (#VC v)
                                      }
                                    , (NONE : thm option)))

                           (L.get_VCs ())
                )


   fun reduce progdef_thm
        =
        let
        val prg = (buildPROG o concl o SPEC_ALL) progdef_thm

        val r = #prparams prg
        val v = #pvparams prg

        val vcsmain = ("MainSpec", 
                       VCG (buildINSTRtree (#pbody prg))

                           ((#ppost prg) <- (mk_olds v // v))

                           (#ppre prg)

                           (fn vc => vc <- (r @ v // mk_olds (r @ v)))
                      )
      in
      VCs := [vcsmain]
      end 

   (* --- a function to verify a single vc --- *)
   fun verify1 (groupNr,vcNr) tactic 
          =
          let
          val vc  = (#VC o fst o List.nth) (snd (List.nth (!VCs,groupNr)), vcNr)
          val thm = prove(vc,tactic)


          fun vcsUpdate ((i,j),f) 
              =
              (VCs := listUpdate (i, (fn (gn,s) => (gn, listUpdate (j,f) s))) (!VCs))
   
          fun updateF (vc,_) = (vc,SOME thm)

          in 
          vcsUpdate ((groupNr,vcNr),updateF) 
          end
 
   (* --- a function to apply a list of lists of tactics to verify
          the entire set of vcs
   --- *)

   fun verify tacticss (* a list of lists of tactic *)
       =
       let
       fun verifyOne (vc,tactic)
          =
          (vc, (SOME (prove(#VC vc,tactic)) handle _ =>  NONE))

       fun verifyGroup ((gname,vcs), tactics)
           =
           (gname, zipWith verifyOne (List.map fst vcs) tactics)
       in
       (VCs := zipWith verifyGroup (!VCs) tacticss ;
        status() )
       end 

   (* --- a function to apply a single (automatic) tactic to verify
          the entire set of vcs
   --- *)

   fun autoverify tactic (* just a single tactic for all vcs *)
       =
       let
       fun verifyOne vc
          =
          (vc, (SOME (prove(#VC vc,tactic)) handle _ =>  NONE))

       fun verifyGroup (gname,vcs)
           =
           (gname, List.map (verifyOne o fst) vcs)
       in
       (VCs := List.map verifyGroup (!VCs)  ;
        status() )
       end 
   
   end 
   end ;

end ;


