(* =================================================================

   This file defines an extension to L0, for writing "unit-test suites".

   A "unit-test suite" is essentially a collection of individual test
   scripts intended to test a particular (L0) program.
   
   An individual test-script is basically just an L0 instruction in which
   the tested program is called. The test-script will typically contain
   "assert" instructions to test that certain conditions are met during
   the test.

   A test-script "S" can be either translated to an executable and then
   run as usual, or reduced to verification conditions and verified.

      (1) When translating to executable:

          * only "assert e" in which e is an Expr (and not an assertion!)
          will be translated to an executable code-frament. If e
          is an asserting, it will be skipped, but a warning will be
          issued during the translation.

          * variable occuring free in S will get random initial value

      (2) When reducing to verification condition:

          * We verify this: {T} S {T}

          * Calls to the tested program will be replaced by the
            program's body (white-box expansion)

            Key-word "whitebox" will be used to indicate that
            white-box expansion should take place.

   A "test-suite" is a collection of test-script, organized 
   in a tree-like structure.

   
   THE GRAMMAR

   // Test-suite

   Test-Suite  -->  TEST  Label?  Test-instruction

                  | SUITE Label? /{ Test-Suite ; ... ; Test-Suite  /}

   Label --> /:: Identifier.   // test/suite name

   Test-instruction --> as L0-instruction, but also allows:

                         whitebox (/@ ProgName(ActualParams))
                         whitebox (Expr /@= ProgName(ActualParams))
   
                         This means that for the verification the 
                         called program is to be treated as a
                         white-box. It means that during the VC-reduction
                         the call will be replaced by the program's
                         body.

                         When we write a test-script for program P,
                         then P should be called as a white-box in
                         the script.

   Author: Wishnu Prasetya
                         
================================================================= *)

local

open Parse HolKernel boolLib bossLib 
open L0Theory

in


val _ = new_theory "Test" ;

val _ = Hol_datatype `TEST = TEST of INSTR | SUITE of TEST` ;

(*--- label *)

val _ = new_binder ("/::", Type `:('a->'a)->'a`) ;

(*--- white-box directive *)

val _ = new_constant ("whitebox", Type `:INSTR->INSTR`) ;

val _ = export_theory();

(* =================================================================

 Example. Assume the following L0 program:

 Define 

      `swap (REF x,REF y) 
          =
          pre  T
          post (x = old y) /\ (y = old x)
          do 
          let tmp = 0
          in /{ tmp /:= x   ;
                x   /:= y   ;
                y   /:= tmp   /}

          return void
      ` ;

  Now we can write the following test-suite in HOL:

  val test0 = --`SUITE /:: main.

                 /{ TEST /:: one. let x=0 and y=1 
                                  in 
                                  /{ whitebox (/@ swap(REF x, REF y))
                                   ; assert((x=1) /\ (y=0))
                                  /}
                 ; TEST /:: two. /{  whitebox (/@ swap(REF x, REF y))
                                   ; whitebox (/@ swap(REF x, REF y))
                                   ; assert(x=y)
                                  /}
                 /}`-- ;


================================================================= *)

end ; 
   
